// Copyright 2012 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// reduq December 2012 version

#ifndef REDUQ_REDUCIBLE_QUEUE_H_
#define REDUQ_REDUCIBLE_QUEUE_H_

#include <algorithm>
#include <limits>
#include <list>
#include <utility>

namespace reduq {

template<typename Reducer>
class reducible_queue {
 private:
  typedef typename Reducer::Type T;
  typedef std::list<std::pair<T, T> > Container;

 public:
  reducible_queue();
  bool empty() const { return container_.empty(); }
  typename Container::size_type size() const { return container_.size(); }
  const T& front() const { return container_.front().first; }
  const T& back() const { return container_.back().first; }
  void push(const T& x) {
    container_.push_back(std::make_pair(x, x));
    T& batch = j_ == container_.end() ? previous_batch_ : current_batch_;
    batch = Reducer::reduce(batch, x);
    Step();
  }
  void pop() {
    container_.pop_front();
    SetJ(j_);
    if (!empty()) Step();
  }
  T reduce() const {
    return (empty() ?
            Reducer::id() :
            Reducer::reduce(container_.front().second,
                            Reducer::reduce(previous_batch_, current_batch_)));
  }

 private:
  void SetJ(typename Container::iterator j) {
    if (j == container_.begin()) {
      j = container_.end();
      previous_batch_ = current_batch_;
      current_batch_ = Reducer::id();
    }
    j_ = j;
  }
  void Step();
  Container container_;
  typename Container::iterator j_;
  T previous_batch_;
  T current_batch_;
  reducible_queue(const reducible_queue&);
  void operator=(const reducible_queue&);
};

template<typename Reducer>
reducible_queue<Reducer>::reducible_queue()
    : container_(),
      j_(container_.end()),
      previous_batch_(Reducer::id()),
      current_batch_(Reducer::id()) {
}

template<typename Reducer>
void reducible_queue<Reducer>::Step() {
  typename Container::iterator i(j_);
  --i;
  if (j_ != container_.end()) i->second = Reducer::reduce(i->first, j_->second);
  SetJ(i);
}

template<typename T>
struct BitwiseAnd {
  typedef T Type;
  static T reduce(const T& x, const T& y) { return x & y; }
  static T id() { return ~T(0); }
};

template<typename T>
struct BitwiseOr {
  typedef T Type;
  static T reduce(const T& x, const T& y) { return x | y; }
  static T id() { return T(0); }
};

template<typename T>
struct BitwiseXor {
  typedef T Type;
  static T reduce(const T& x, const T& y) { return x ^ y; }
  static T id() { return T(0); }
};

template<typename T>
struct IntegerMax {
  typedef T Type;
  static T reduce(const T& x, const T& y) { return std::max(x, y); }
  static T id() { return std::numeric_limits<T>::min(); }
};

template<typename T>
struct IntegerMin {
  typedef T Type;
  static T reduce(const T& x, const T& y) { return std::min(x, y); }
  static T id() { return std::numeric_limits<T>::max(); }
};

template<typename T>
struct Product {
  typedef T Type;
  static T reduce(const T& x, const T& y) { return x * y; }
  static T id() { return T(1); }
};

template<typename T>
struct RealMax {
  typedef T Type;
  static T reduce(const T& x, const T& y) { return std::max(x, y); }
  static T id() { return -std::numeric_limits<T>::infinity(); }
};

template<typename T>
struct RealMin {
  typedef T Type;
  static T reduce(const T& x, const T& y) { return std::min(x, y); }
  static T id() { return std::numeric_limits<T>::infinity(); }
};

template<typename T>
struct Sum {
  typedef T Type;
  static T reduce(const T& x, const T& y) { return x + y; }
  static T id() { return T(0); }
};
}  // namespace reduq
#endif  // REDUQ_REDUCIBLE_QUEUE_H_
